/*=========================================================================

  Program:   Visualization Toolkit
  Module:    vtkFeatures.h.in

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#ifndef vtkFeatures_h
#define vtkFeatures_h

/* Should all New methods use the object factory override. */
/* #undef VTK_ALL_NEW_OBJECT_FACTORY */

/* Whether VTK supports an extended memory space via memkind. */
/* #undef VTK_USE_MEMKIND */

#endif
