/*=========================================================================

  Program:   Visualization Toolkit

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/

#ifndef vtkCompositePolyDataMapper2Internal_h
#define vtkCompositePolyDataMapper2Internal_h

#include "vtkDeprecation.h" // for deprecation macros

#if VTK_DEPRECATION_LEVEL >= VTK_VERSION_CHECK(9, 1, 0)
#ifdef _MSC_VER
#pragma message(                                                                                   \
  "This header has been renamed to `vtkCompositeMapperHelper2.h` to match its provided class name")
#else
#warning                                                                                           \
  "This header has been renamed to `vtkCompositeMapperHelper2.h` to match its provided class name"
#endif
#endif

#include "vtkCompositeMapperHelper2.h"

#endif

// VTK-HeaderTest-Exclude: vtkCompositePolyDataMapper2Internal.h
